
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2_arserver:arcor2_arserver_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 - Augmented Reality Collaborative Robot.',
    'license': 'LGPL',
    'long_description': """# arcor2_arserver
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),

## [0.10.1] - 2020-10-30

### Changed
- RPC for IK/FK.

## [0.10.0] - 2020-10-22

### Changed
- Updated due to a breaking change in `arcor2_execution_data`.

## [0.9.1] - 2020-10-19

### Changed
- ARCOR2 dependency updated

### Fixed
- fixed check for existing override
 
## [0.9.0] - 2020-10-16

### Changed
- WS API for updates of scene objects parameters and management of project-defined overrides.
- Integration tests refactored into multiple modules that can run in parallel.
- `UpdateObjectPose` and `UpdateActionPointPosition` now has dry_run.
- Box/Cylinder/Sphere models now have some constraints on their dimensions.
- 

## [0.8.0] - 2020-09-24
### Changed
- The first release of the separated package.
- RPC/Events refactored. 
- ARServer script now has ```--openapi``` argument and is able to generate OpenAPI models for RPCs and events.
- Detection of loops in: AP hierarchy, logic.
- Caching client for Project service.
- Proper handling of object inheritance.
- ObjectTypes are stored in temporary module on disk.
- Integration tests.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2_arserver',
    'python_requires': '==3.8.*',
    'version': '0.10.1',
    'package_dir': {
        '': 'src',
    },
    'packages': (
        'arcor2_arserver',
        'arcor2_arserver.clients',
        'arcor2_arserver.object_types',
        'arcor2_arserver.rpc',
        'arcor2_arserver.scripts',
    ),
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2_arserver': (
            'VERSION',
            'py.typed',
        ),
    },
    'install_requires': (
        'Flask==1.1.2',
        'PyYAML==5.3.1',
        'aiologger==0.6.0',
        'aiorun==2020.6.1',
        'apispec-webframeworks==0.5.2',
        'apispec==3.3.2',
        'arcor2_arserver_data~=0.9.2',
        'arcor2_build_data~=0.8.1',
        'arcor2_calibration_data~=0.1.0',
        'arcor2_execution_data~=0.9.0',
        'arcor2~=0.9.2',
        'dataclasses-jsonschema-rereleased[fast-validation]==2.12.2',
        'dataclasses==0.6',
        'lru-dict==1.1.6',
        'numpy-quaternion==2020.9.5.14.42.2',
        'pydobot2==0.1.0',
        'pyhumps==1.6.1',
        'typed-ast==1.4.1',
        'typing-inspect==0.6.0',
        'websockets==8.1',
    ),
    'entry_points': {
        'console_scripts': [
            'arcor2_arserver=arcor2_arserver.scripts.arserver:main',
            'arcor2_broadcaster=arcor2_arserver.scripts.broadcaster:main',
        ],
    },
})
